import os
import cv2
import random
import numpy as np
import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
from collections import deque
from gameAPI.agent import gameAgent


class REINFORCE:
    def __init__(self, config, **kwargs):
        self.config = config
        self.policy_net = self._create_network()

    def _create_network(self):
        # 创建策略网络
        x = tf.placeholder(tf.float32, [None, *self.config.frame_size, 3])
        
        # 网络层的示例，这应根据你的需求进行调整
        conv1 = tf.layers.conv2d(inputs=x, filters=32, kernel_size=[8, 8], strides=4, activation=tf.nn.relu)
        conv2 = tf.layers.conv2d(inputs=conv1, filters=64, kernel_size=[4, 4], strides=2, activation=tf.nn.relu)
        conv3 = tf.layers.conv2d(inputs=conv2, filters=64, kernel_size=[3, 3], strides=1, activation=tf.nn.relu)
        flattened = tf.layers.flatten(conv3)
        fc = tf.layers.dense(flattened, 512, activation=tf.nn.relu)
        action_probs = tf.layers.dense(fc, self.config.num_actions, activation=tf.nn.softmax)

        return x, action_probs

    def train(self, session):
        # create network
        net_in, action_probs = self._create_network()  # 策略网络输出动作概率
        actions_placeholder = tf.placeholder(tf.int32, [None])
        rewards_placeholder = tf.placeholder(tf.float32, [None])

        # 获取选中动作的概率
        action_masks = tf.one_hot(actions_placeholder, 3)  # 假设有3个动作
        picked_action_probs = tf.reduce_sum(action_probs * action_masks, axis=1)

        # 损失函数和训练步骤
        loss = -tf.reduce_mean(tf.log(picked_action_probs) * rewards_placeholder)
        train_step = tf.train.AdamOptimizer(self.config.lr).minimize(loss)

        saver = tf.train.Saver()
        session.run(tf.global_variables_initializer())

        # game agent
        game_agent = gameAgent()

        # 训练循环
        for episode in range(self.config.num_episodes):
            frame_info = game_agent.nextFrame(None)
            state = frame_info[0]  # 提取 frame
            state = cv2.resize(state, (96, 96))  # 调整 state 大小
            episode_rewards = []
            episode_actions = []
            episode_states = []
            done = False
            while not done:
                # 获取动作概率
                action_probabilities = session.run(action_probs, feed_dict={net_in: [state]})
                action_index = np.random.choice(range(3), p=action_probabilities[0])

                # 将动作索引转换为所需格式
                action = [0, 0, 0]
                action[action_index] = 1


                frame_info = game_agent.nextFrame([1 if i == action_index else 0 for i in range(3)])
                next_state = frame_info[0]  # 提取 next frame
                next_state = cv2.resize(next_state, (96, 96))  # 调整 next_state 大小
                
                reward = frame_info[2]  # 提取 reward
                done = frame_info[3]  # 提取 terminal

                episode_states.append(state)
                episode_actions.append(action_index)
                episode_rewards.append(reward)

                state = next_state 

            # 计算累积回报
            cumulative_rewards = self._compute_cumulative_rewards(episode_rewards)

            # 更新策略
            session.run(train_step, feed_dict={
                net_in: episode_states,
                actions_placeholder: episode_actions,
                rewards_placeholder: cumulative_rewards
            })

        if episode % self.config.save_interval == 0:
            saver.save(session, os.path.join(self.config.model_dir, 'REINFORCE_Pong'), global_step=episode)


    def _compute_cumulative_rewards(self, rewards):
        # 计算累积回报
        cumulative_rewards = np.zeros_like(rewards, dtype=np.float32)
        running_add = 0
        for t in reversed(range(len(rewards))):
            running_add = running_add * self.config.gamma + rewards[t]
            cumulative_rewards[t] = running_add
        return cumulative_rewards